#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import pytest
import rich.errors
import rich.style
import tomlkit

from dye.exceptions import DyeError, DyeSyntaxError
from dye.pattern import Pattern
from dye.scope import Scope
from dye.theme import Theme
from dye.utils import DefinitionSource

SAMPLE_THEME = """
[colors]
foreground = "#f8f8f2"
foreground_high = "foreground"
foreground_medium = "foreground"
foreground_low = "foreground"

triad.first = "#aaaaaa"
triad.second = "#bbbbbb"
triad.third = "#cccccc"

[styles]
themeonly = "{{color.foreground}} on #393b47"
foreground = "{{ color.foreground }}"
"""

SAMPLE_PATTERN = """
description = "Oxygen is a pattern with lots of space"
type = "dark"
version = "2.0"

requires_theme = "reqtheme"
prevent_themes = true

[colors]
pattern_purple =  "#bd93f8"
pattern_yellow =  "#f1fa8b"
notyet =  "{{ colors.yellow }}"

foreground =  "#e9e9e3"
background =  "#282a36"

foreground_unknown = "{{ colors.unknown }}"

background_high = "{{ colors.background }}"
background_medium = "{{ color.background }}"
background_low = "background"
background_double1 = "{{ color.background_low }}"
background_double2 = "background_medium"

yellow = "#e2eb9c"

triad.second = "#dd2222"

tetrad.first = "#ff1111"
tetrad.second = "#ff2222"
tetrad.third = "foreground_low"
tetrad.fourth = "{{ color.pattern_purple }}"
tetrad.fifth = "{{ colors.triad.third }}"

group1.1 = "triad.first"
group1.2 = "{{ color.triad.second }}"
group1.3 = "{{ colors.triad.third }}"


[styles]
notyet = "{{ styles.text }}"
foreground = "{{ color.foreground }}"
text = "bold {{ colors.foreground }} on {{ colors.background }}"
text_high = "{{ styles.text }}"
text_medium = "{{ style.text }}"
text_low = "text"
text_double1 = "{{ style.text_low }}"
text_double2 = "text_medium"
text_colorref = "{{ color.foreground_medium }}"
color1 = "#ff79c6"
color2 = "{{ colors.unknown }}"
color3 = "{{ style.unknown }}"
color4 = ""

pattern_text = '#cccccc on #ffffff'
pattern_text_high = '#000000 on #ffffff'
pattern_text_low = '#999999 on #ffffff'
pattern_yellow = "{{ colors.pattern_yellow }}"

triad_sty.first = "{{ colors.triad.first }}"
triad_sty.second = "pattern_text"
triad_sty.third = "{{ style.themeonly }}"

[variables]
capture.somevar = "printf '%s' jojo"
secondhalf = "5555"
replace = "{{variables.secondhalf}}"
firsthalf = "fred"
myred = "{{variables.firsthalf}}{{variables.secondhalf}}"
v_yellow = "{{styles.pattern_yellow|fg_hex_no_hash}}"
capture.anothervar = "printf '%s' {{colors.pattern_purple}}"

[scopes.iterm]
agent = "iterm"
cursor = "block"

[scopes.nocolor]
agent = "environment_variables"
export.NO_COLOR = "true"

[scopes.shell]
agent = "shell"
is_enabled = false
command.dontrun = "echo qqq"
"""


@pytest.fixture
def sthm():
    """the sample theme"""
    return Theme.loads(SAMPLE_THEME)


@pytest.fixture
def spat():
    """the sample pattern without loading the theme"""
    pattern = Pattern.loads(SAMPLE_PATTERN)
    return pattern


@pytest.fixture
def sthmpat():
    """the sample pattern with the sample theme merged into it"""
    theme = Theme.loads(SAMPLE_THEME)
    pattern = Pattern.loads(SAMPLE_PATTERN, theme)
    return pattern


#
# test dunders
#
def test_truthy1(spat):
    assert spat


def test_truthy2(sthmpat):
    assert sthmpat


def test_falsy1():
    assert not Pattern()


def test_falsy2():
    assert not Pattern.loads("")


def test_falsy3():
    assert not Pattern.loads(None)


#
# make sure load() and loads() work properly
#
def test_load(tmp_path):
    # def test_load_from_args_theme_name(dye, mocker, tmp_path):
    # give a theme name, but the full name including the .toml
    patternfile = tmp_path / "pattern.toml"
    with open(patternfile, "w", encoding="utf8") as fvar:
        fvar.write(SAMPLE_PATTERN)

    with open(patternfile, encoding="utf8") as fvar:
        pat = Pattern.load(fvar)
        assert pat.definition


def test_loads(spat):
    assert isinstance(spat.definition, dict)
    assert spat.definition


def test_loads_empty():
    pat = Pattern.loads("")
    assert pat.definition == {}


def test_loads_none():
    pat = Pattern.loads(None)
    assert pat.definition == {}


def test_loads_colors(spat):
    assert isinstance(spat.colors, dict)
    assert isinstance(spat.colors["pattern_purple"], str)
    assert spat.colors["pattern_purple"] == "#bd93f8"


def test_loads_styles(spat):
    assert isinstance(spat.styles, dict)
    assert isinstance(spat.styles["pattern_text"], rich.style.Style)
    assert isinstance(spat.styles["pattern_text_high"], rich.style.Style)


#
# test pattern metadata/properties
# these tests just ensure the data is extracted propertly
# from the toml
#
def test_description(spat):
    assert spat.description == "Oxygen is a pattern with lots of space"


def test_no_description():
    pattern_str = """prevent_themes = true"""
    pat = Pattern.loads(pattern_str)
    assert not pat.description


def test_type_not_defined():
    pattern_str = """
    [colors]
    foreground = "#d2d2d2"
    """
    pattern = Pattern.loads(pattern_str)
    assert not pattern.type


def test_type():
    pattern_str = """
    type = "dark"
    """
    pattern = Pattern.loads(pattern_str)
    assert pattern.type == "dark"


def test_version_not_defined():
    pattern_str = """
    [colors]
    foreground = "#d2d2d2"
    """
    pattern = Pattern.loads(pattern_str)
    assert not pattern.version


def test_version():
    pattern_str = """
    version = "2.1"
    """
    pattern = Theme.loads(pattern_str)
    assert pattern.version == "2.1"


def test_prevent_themes():
    pattern_str = """prevent_themes = true"""
    pat = Pattern.loads(pattern_str)
    assert pat.prevent_themes


def test_prevent_themes_not_defined():
    pattern_str = """description = 'hi'"""
    pat = Pattern.loads(pattern_str)
    assert not pat.prevent_themes


def test_prevent_themes_not_boolean():
    pattern_str = "prevent_themes = 'nope'"
    with pytest.raises(DyeSyntaxError):
        Pattern.loads(pattern_str)


def test_requires_theme():
    pattern_str = """requires_theme = '/path/to/theme'"""
    pat = Pattern.loads(pattern_str)
    assert pat.requires_theme == "/path/to/theme"


def test_requires_theme_not_defined():
    pattern_str = """description = 'hi'"""
    pat = Pattern.loads(pattern_str)
    assert not pat.requires_theme


#
# test processing of colors
#
def test_color_pattern(sthmpat):
    """a color that is only defined in the pattern"""
    assert sthmpat.colors["background"] == "#282a36"


def test_color_theme(sthmpat):
    """a color that is only defined in the theme"""
    assert sthmpat.colors["foreground_high"] == "#f8f8f2"


def test_colors_must_be_strings():
    pattern_str = """
    [colors]
    background = 282
    """
    with pytest.raises(DyeSyntaxError):
        Pattern.loads(pattern_str)


def test_color_pattern_override_theme(sthm, spat, sthmpat):
    """a color defined in both pattern and theme, make sure pattern overrides"""
    assert sthm.colors["foreground"] == "#f8f8f2"
    # make sure we get the same answer whether we load the theme or not
    assert spat.colors["foreground"] == "#e9e9e3"
    assert sthmpat.colors["foreground"] == "#e9e9e3"


def test_colors_reference(sthmpat):
    assert sthmpat.colors["background_high"] == sthmpat.colors["background"]


def test_color_reference(sthmpat):
    assert sthmpat.colors["background_medium"] == sthmpat.colors["background"]


def test_colors_bare_reference(sthmpat):
    assert sthmpat.colors["background_low"] == sthmpat.colors["background"]


def test_colors_double1_reference(sthmpat):
    assert sthmpat.colors["background_double1"] == sthmpat.colors["background"]


def test_colors_double2_reference(sthmpat):
    assert sthmpat.colors["background_double2"] == sthmpat.colors["background"]


def test_colors_unknown_reference(sthmpat):
    assert sthmpat.colors["foreground_unknown"] == ""


def test_colors_load_order(sthmpat):
    assert sthmpat.colors["notyet"] == ""


def test_colors_subtable(sthmpat):
    assert sthmpat.colors["tetrad.first"] == "#ff1111"
    assert sthmpat.colors["tetrad.second"] == "#ff2222"


def test_colors_subtable_from_theme(sthmpat):
    assert sthmpat.colors["triad.first"] == "#aaaaaa"
    assert sthmpat.colors["triad.second"] == "#dd2222"
    assert sthmpat.colors["triad.third"] == "#cccccc"


def test_colors_subtable_reference1(sthmpat):
    # theme colors get resolved to values before pattern
    # colors do. Therefore, foreground_low for tetrad.third
    # resolves to #f8f8f2 not #e9e9e3
    assert sthmpat.colors["tetrad.third"] == "#f8f8f2"


def test_colors_subtable_reference2(sthmpat):
    assert sthmpat.colors["tetrad.fourth"] == sthmpat.colors["pattern_purple"]


def test_colors_subtable_reference3(sthmpat):
    assert sthmpat.colors["tetrad.fifth"] == "#cccccc"


def test_colors_subtable_reference4(sthmpat):
    assert sthmpat.colors["group1.1"] == sthmpat.colors["triad.first"]


def test_colors_subtable_reference5(sthmpat):
    assert sthmpat.colors["group1.2"] == sthmpat.colors["triad.second"]


def test_colors_subtable_reference6(sthmpat):
    assert sthmpat.colors["group1.3"] == sthmpat.colors["triad.third"]


#
# test processing of styles
#
def test_empty_pattern(sthm):
    pattern = Pattern.loads("", sthm)
    assert pattern.colors["foreground"]
    assert pattern.styles["themeonly"]


def test_style_pattern(spat):
    """defined only in the pattern, referencing colors
    only in the pattern"""
    assert spat.styles["text"].color.name == "#e9e9e3"
    assert spat.styles["text"].color.triplet.hex == "#e9e9e3"
    assert spat.styles["text"].bgcolor.name == "#282a36"
    assert spat.styles["text"].bgcolor.triplet.hex == "#282a36"


def test_styles_must_be_strings():
    pattern_str = """
    [styles]
    text = 282
    """
    with pytest.raises(DyeSyntaxError):
        Pattern.loads(pattern_str)


def test_style_theme(sthmpat):
    """a style that is only defined in the theme"""
    assert sthmpat.styles["themeonly"].bgcolor.name == "#393b47"


def test_style_pattern_overrides_theme(sthm, spat, sthmpat):
    """a color defined in both pattern and theme, make sure pattern overrides"""
    # here's the definition from the theme
    assert sthm.styles["foreground"].color.name == "#f8f8f2"
    # make sure we get the same answer whether we load the theme or not
    assert spat.styles["foreground"].color.name == "#e9e9e3"
    assert sthmpat.styles["foreground"].color.name == "#e9e9e3"


def test_style_reference_color_only_in_theme(sthmpat):
    assert sthmpat.styles["text_colorref"].color.name == "#f8f8f2"


def test_style_color_ref(sthmpat):
    """a style that references a color"""
    assert sthmpat.styles["foreground"].color.name == sthmpat.colors["foreground"]


def test_style_no_color_reference(sthmpat):
    """test a style that doesn't reference any colors, just a plain
    hexcode as the definition"""
    assert sthmpat.styles["color1"].color.name == "#ff79c6"
    assert sthmpat.styles["color1"].color.triplet.hex == "#ff79c6"


def test_styles_reference(sthmpat):
    """a style that references another style as {{styles.text_high}}"""
    assert sthmpat.styles["text_high"] == sthmpat.styles["text"]


def test_style_reference(sthmpat):
    """a style that references another style as {{style.text_medium}}"""
    assert sthmpat.styles["text_medium"] == sthmpat.styles["text"]


def test_styles_bare_reference(sthmpat):
    assert sthmpat.styles["text_low"] == sthmpat.styles["text"]


def test_styles_double1_reference(sthmpat):
    assert sthmpat.styles["text_double1"] == sthmpat.styles["text"]


def test_styles_double2_reference(sthmpat):
    assert sthmpat.styles["text_double2"] == sthmpat.styles["text"]


def test_style_unknown_color_reference(sthmpat):
    # look for a style object, but one that's empty
    assert isinstance(sthmpat.styles["color2"], rich.style.Style)
    assert not sthmpat.styles["color2"]


def test_style_unknown_style_reference(sthmpat):
    # look for a style object, but one that's empty
    assert isinstance(sthmpat.styles["color3"], rich.style.Style)
    assert not sthmpat.styles["color3"]


def test_styles_load_order(sthmpat):
    # look for a style object, but one that's empty
    assert isinstance(sthmpat.styles["notyet"], rich.style.Style)
    assert not sthmpat.styles["notyet"]


def test_style_empty(sthmpat):
    assert isinstance(sthmpat.styles["color4"], rich.style.Style)
    assert not sthmpat.styles["color4"]


def test_styles_subtable(sthmpat):
    assert isinstance(sthmpat.styles["triad_sty"], dict)
    assert sthmpat.styles["triad_sty.first"].color.name == "#aaaaaa"


def test_styles_subtable_reference1(sthmpat):
    assert sthmpat.styles["triad_sty.second"].color.name == "#cccccc"


def test_styles_subtable_reference2(sthmpat):
    assert sthmpat.styles["triad_sty.third"].bgcolor.name == "#393b47"


#
# test variable definition and usage in the [variables] table
#
def test_variable():
    pattern_str = """
        [variables]
        varname = "value"
    """
    pattern = Pattern.loads(pattern_str)
    assert pattern.variables["varname"] == "value"


def test_variables_reference():
    pattern_str = """
        [variables]
        varname = "value"
        var1 = "{{var.varname}}"
        var2 = "{{vars.varname}}"
        var3 = "{{variable.varname}}"
        var4 = "{{variables.varname}}"
    """
    pattern = Pattern.loads(pattern_str)
    assert pattern.variables["var1"] == "value"
    assert pattern.variables["var2"] == "value"
    assert pattern.variables["var3"] == "value"
    assert pattern.variables["var4"] == "value"


def test_undefined_variable_reference():
    pattern_str = """
        [variables]
        varname = "value"
        var1 = "{{var.notdefined}}"
    """
    pattern = Pattern.loads(pattern_str)
    assert pattern.variables["var1"] == ""


def test_variables_color_reference():
    pattern_str = """
        [colors]
        bright_green = "#99e343"

        [variables]
        var1 = "{{color.bright_green}}"
        var2 = "{{colors.bright_green}}"
    """
    pattern = Pattern.loads(pattern_str)
    assert pattern.variables["var1"] == "#99e343"
    assert pattern.variables["var1"] == "#99e343"


def test_variables_style_reference():
    pattern_str = """
        [styles]
        bright_green = "#99e343"

        [variables]
        var1 = "{{style.bright_green|fg_hex}}"
        var2 = "{{styles.bright_green|fg_hex}}"
    """
    pattern = Pattern.loads(pattern_str)
    assert pattern.variables["var1"] == "#99e343"
    assert pattern.variables["var1"] == "#99e343"


def test_capture_variable():
    pattern_str = """
        [variables]
        capture.somevar = "printf 'hello there'"
    """
    pattern = Pattern.loads(pattern_str)
    assert pattern.variables["somevar"] == "hello there"


def test_capture_variable_error():
    pattern_str = """
        [variables]
        capture.somevar = "barf_is_not_a_shell_command"
    """
    with pytest.raises(DyeError):
        Pattern.loads(pattern_str)


def test_variable_redefine1_error():
    pattern_str = """
        [variables]
        somevar = "builtin echo hi"
        somevar = "can't do this"
    """
    with pytest.raises(tomlkit.exceptions.KeyAlreadyPresent):
        Pattern.loads(pattern_str)


def test_variable_redefine2_error():
    pattern_str = """
        [variables]
        capture.somevar = "builtin echo hi"
        somevar = "can't do this"
    """
    with pytest.raises(DyeError):
        Pattern.loads(pattern_str)


#
# test a few scope related things
#
def test_has_scope():
    pattern_str = """
        [scopes.qqq]
        agent = "iterm"
        style.foreground = "blue"
        style.background = "white"
    """
    pattern = Pattern.loads(pattern_str)

    assert pattern.has_scope("qqq")
    assert not pattern.has_scope("fred")


def test_scopes():
    pattern_str = """
        [scopes.qqq]
        agent = "iterm"
        style.foreground = "blue"
        style.background = "white"

        [scopes.fff]
        agent = "environment_variables"
        unset = "listvar"
    """
    pattern = Pattern.loads(pattern_str)

    assert len(pattern.scopes) == 2
    assert isinstance(pattern.scopes["qqq"], Scope)
    assert pattern.scopes["fff"].name == "fff"


def test_scopes_empty():
    pattern_str = """
        [variables]
        hi = "there"
    """
    pattern = Pattern.loads(pattern_str)

    assert len(pattern.scopes) == 0
    assert pattern.scopes == {}


#
# test other methods
#
COLOR_DEFS = [
    ("foreground_high", DefinitionSource.THEME, "foreground"),
    ("pattern_purple", DefinitionSource.PATTERN, "#bd93f8"),
    ("triad.first", DefinitionSource.THEME, "#aaaaaa"),
    ("triad.second", DefinitionSource.PATTERN, "#dd2222"),
]


@pytest.mark.parametrize("color_name, expected_source, expected_defn", COLOR_DEFS)
def test_get_color_def_theme(sthmpat, color_name, expected_source, expected_defn):
    (source, defn) = sthmpat.get_color_def(color_name)
    assert source == expected_source
    assert defn == expected_defn


def test_get_color_def_notfound(sthmpat):
    with pytest.raises(KeyError):
        sthmpat.get_color_def("notfoundhere")


STYLE_DEFS = [
    ("themeonly", DefinitionSource.THEME, "{{color.foreground}} on #393b47"),
    ("pattern_text_low", DefinitionSource.PATTERN, "#999999 on #ffffff"),
]


@pytest.mark.parametrize("style_name, expected_source, expected_defn", STYLE_DEFS)
def test_get_style_def_theme(sthmpat, style_name, expected_source, expected_defn):
    (source, defn) = sthmpat.get_style_def(style_name)
    assert source == expected_source
    assert defn == expected_defn


def test_get_style_def_notfound(sthmpat):
    with pytest.raises(KeyError):
        sthmpat.get_style_def("notfoundhere")



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""class for storing and processing a pattern"""

import contextlib
import subprocess

import jinja2
import tomlkit
from benedict import benedict

from .exceptions import DyeError, DyeSyntaxError
from .filters import jinja_filters
from .scope import Scope
from .utils import DefinitionSource, merge_and_process_colors, merge_and_process_styles


class Pattern:
    """load and parse a pattern file into a pattern object"""

    @staticmethod
    def loads(tomlstring=None, theme=None):
        """Load a pattern from a given string, and return a new pattern object

        doesn't do any processing or applying of the pattern
        """
        if tomlstring:  # noqa: SIM108
            toparse = tomlstring
        else:
            # tomlkit can't parse None, so if we got it as the default
            # or if the caller pased None intentionally...
            toparse = ""
        pattern = Pattern()
        pattern.definition = benedict(tomlkit.loads(toparse))
        pattern.process(theme)
        return pattern

    @staticmethod
    def load(fobj, theme=None, filename=None):
        """Load a pattern a file object

        doesn't do any processing or applying of the pattern
        """
        pattern = Pattern()
        pattern.definition = benedict(tomlkit.load(fobj))
        pattern.filename = filename
        pattern.process(theme)

        return pattern

    #
    # initialization and properties
    #
    def __init__(self):
        """Construct a new Pattern object"""

        # the raw toml definition of the pattern
        self.definition = benedict()

        # if we have applied a theme to the pattern, store it here
        self.theme = None

        # these contain the core parts of the pattern,
        # but they have all been processed through the template
        # so they can be used by consumers of our class.
        # these are all set by process()
        self.colors = benedict()
        self.styles = benedict()
        self.variables = benedict()
        self.scopes = benedict()

        # metadata and pattern info
        self.description = None
        self.type = None
        self.version = None
        self.prevent_themes = False
        self.requires_theme = None

        # a place to stash the file that the pattern was loaded from
        # it's up to the caller/user to make sure this is set properly
        # defaults to None
        self.filename = None

    def __bool__(self):
        """Falsy if the definition is falsy, truthy if the definition is truthy"""
        return bool(self.definition)

    def process(self, theme=None):
        """Process the loaded pattern definition, merging in a theme if given

        returns nothing, populates stuff in the current object:

            .theme
            .colors
            .styles
            .variables
            .scopes
        """
        self.theme = theme

        self.description = self.definition.get("description", None)
        self.type = self.definition.get("type", None)
        self.version = self.definition.get("version", None)

        prevent = self.definition.get("prevent_themes", False)
        if not isinstance(prevent, bool):
            raise DyeSyntaxError("'prevent_themes' must be true or false")
        self.prevent_themes = prevent

        self.requires_theme = self.definition.get("requires_theme", None)

        jinja_env = jinja2.Environment()
        jinja_env.filters = jinja_filters()

        self._process_colors(jinja_env)
        self._process_styles(jinja_env)
        self._process_variables(jinja_env)
        self._process_scopes()

    def _process_colors(self, jinja_env):
        """merge the colors from the pattern and self.theme together

        this sets self.colors

        A color can be referenced in this section in any of these ways:

        foreground = "#f8f8f2"
        foreground_high = "foreground"
        foreground_medium = "{{ colors.foreground }}"
        foreground_low = "{{ color.foreground }}"

        A color is just a string, there is nothing done to the value of the color.
        You could do everything with a variable that you can do with a color,
        it's just a convenient way to group/name them.
        """
        self.colors = self.theme.colors.copy() if self.theme else benedict()
        pattern_colors = benedict()
        with contextlib.suppress(KeyError):
            pattern_colors = benedict(self.definition["colors"])
        merge_and_process_colors(self.colors, pattern_colors, jinja_env)

    def _process_styles(self, jinja_env):
        """merge the styles from this pattern and the given theme together

        this sets self.styles
        """
        self.styles = self.theme.styles.copy() if self.theme else benedict()
        pattern_styles = benedict()
        with contextlib.suppress(KeyError):
            pattern_styles = benedict(self.definition["styles"])
        merge_and_process_styles(self.styles, pattern_styles, jinja_env, self.colors)

    def _process_variables(self, jinja_env):
        """process the variables into self.variables"""
        # Process the capture variables without rendering.
        # We can't render because the toml parser has to group
        # all the "capture" items in a separate table, they can't be
        # interleaved with the regular variables in the order they are
        # defined. So we have to choose to process either the [variables]
        # table or the [variables][capture] table first. We choose the
        # [variables][capture] table.
        #
        # no technical reason why we don't render colors and styles
        # in capture variables
        processed_vars = {}
        try:
            cap_vars = self.definition["variables"]["capture"]
        except KeyError:
            cap_vars = {}
        for var, cmd in cap_vars.items():
            proc = subprocess.run(cmd, shell=True, check=False, capture_output=True)
            if proc.returncode != 0:
                raise DyeError(
                    f"capture variable '{var}' returned a non-zero exit code."
                )
            processed_vars[var] = str(proc.stdout, "UTF-8")

        # then add the regular variables, processing them as templates
        try:
            # make a shallow copy, because we are gonna delete any capture
            # vars and we want the definition to stay pristine
            reg_vars = self.definition["variables"].copy()
        except KeyError:
            reg_vars = {}
        # if no capture variables, we don't care, if present
        # delete that table so we don't process it again
        with contextlib.suppress(KeyError):
            del reg_vars["capture"]

        for var, definition in reg_vars.items():
            if var in processed_vars:
                raise DyeError(f"variable '{var}' has already been defined.")
            template = jinja_env.from_string(definition)
            processed_vars[var] = template.render(
                color=self.colors,
                colors=self.colors,
                style=self.styles,
                styles=self.styles,
                var=processed_vars,
                vars=processed_vars,
                variable=processed_vars,
                variables=processed_vars,
            )

        self.variables = processed_vars

    def _process_scopes(self):
        """process value in every scope as a template to resolve
        colors, styles, and variables

        sets self.scopes as a dict of objects
        """
        raw_scopes = {}
        with contextlib.suppress(KeyError):
            raw_scopes = self.definition["scopes"]

        for name in raw_scopes:
            scope = Scope(name, self)
            self.scopes[name] = scope

    #
    # other methods
    #
    def has_scope(self, scope):
        """Check if the given scope exists."""
        return scope in self.scopes

    def get_color_def(self, color_name):
        """retrieve the definition of color_name, from wherever it was defined

        Args:
            color_name (str): the name of the color you want to retrieve

        Raises:
            KeyError: if color_name is not defined anywhere

        Returns:
            (source, color_def): a DefinitionSource of where the color_name is defined
            and a string of the definition
        """
        found = False
        source = None
        color_def = None

        if self.theme:
            with contextlib.suppress(KeyError):
                color_def = self.theme.definition["colors"][color_name]
                source = DefinitionSource.THEME
                found = True
        if self.definition:
            with contextlib.suppress(KeyError):
                color_def = self.definition["colors"][color_name]
                source = DefinitionSource.PATTERN
                found = True

        if not found:
            raise KeyError(color_name)
        return (source, color_def)

    def get_style_def(self, style_name):
        """retrieve the definition of style_name from theme or pattern"""
        found = False
        source = None
        style_def = None

        if self.theme:
            with contextlib.suppress(KeyError):
                style_def = self.theme.definition["styles"][style_name]
                source = DefinitionSource.THEME
                found = True
        if self.definition:
            with contextlib.suppress(KeyError):
                style_def = self.definition["styles"][style_name]
                source = DefinitionSource.PATTERN
                found = True

        if not found:
            raise KeyError(style_name)
        return (source, style_def)



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import pytest
import rich

from dye import DyeSyntaxError, Theme

SAMPLE_THEME = """
    [colors]
    background =  "#282a36"
    foreground =  "#f8f8f2"
    notyet =  "{{ colors.yellow }}"
    green =  "#50fa7b"
    orange =  "#ffb86c"
    pink =  "#ff79c6"
    purple =  "#bd93f9"
    yellow =  "#f1fa8c"

    foreground_unknown = "{{ colors.unknown }}"

    background_high = "{{ colors.background }}"
    background_medium = "{{ color.background }}"
    background_low = "background"
    background_double1 = "{{ color.background_low }}"
    background_double2 = "background_medium"

    triad.first = "purple"
    triad.second = "yellow"
    triad.third = "green"

    tetrad.first = "triad.first"
    tetrad.second = "yellow"
    tetrad.third = "pink"
    tetrad.fourth = "{{ colors.triad.third }}"

    accent.1 = "orange"
    accent.two = "pink"

    [styles]
    notyet = "{{ styles.foreground }}"
    foreground = "{{ color.foreground }}"
    text = "bold {{ colors.foreground }} on {{ colors.background }}"
    text_high = "{{ styles.text }}"
    text_medium = "{{ style.text }}"
    text_low = "text"
    text_double1 = "{{ style.text_low }}"
    text_double2 = "text_medium"
    color1 = "#ff79c6"
    color2 = "{{ colors.unknown }}"
    color3 = ""

    triad.first = "{{ color.green }} on {{ color.pink }}"
    triad.second = "foreground"
    triad.third = "{{ style.triad.first }}"

    accent.1 = "{{ color['accent.1'] }}"
    accent.2 = "{{ color['accent.two'] }}"
"""


@pytest.fixture
def sthm():
    return Theme.loads(SAMPLE_THEME)


#
# test dunders
#
def test_truthy(sthm):
    assert sthm


def test_falsy1():
    assert not Theme()


def test_falsy2():
    assert not Theme.loads("")


def test_falsy3():
    assert not Theme.loads(None)


#
# test load() and loads()
#
def test_load(tmp_path):
    # def test_load_from_args_theme_name(dye, mocker, tmp_path):
    # give a theme name, but the full name including the .toml
    themefile = tmp_path / "oxygen.toml"
    with open(themefile, "w", encoding="utf8") as fvar:
        fvar.write(SAMPLE_THEME)

    with open(themefile, encoding="utf8") as fvar:
        theme = Theme.load(fvar, filename=themefile)
    # Theme.load() uses the same code as Theme.loads(), so we don't
    # have to retest everything. If loads() works and load() can
    # open and read the file, load() will work too
    assert isinstance(theme.definition, dict)
    assert len(theme.definition) == 2


def test_loads(sthm):
    assert isinstance(sthm.definition, dict)


def test_loads_colors(sthm):
    assert isinstance(sthm.colors, dict)
    assert isinstance(sthm.colors["orange"], str)
    assert sthm.colors["orange"] == "#ffb86c"


def test_loads_styles(sthm):
    assert isinstance(sthm.styles, dict)
    assert isinstance(sthm.styles["text"], rich.style.Style)
    assert isinstance(sthm.styles["text_high"], rich.style.Style)
    assert isinstance(sthm.styles["color1"], rich.style.Style)


def test_loads_empty():
    theme = Theme.loads("")
    assert not theme
    assert theme.definition == {}
    assert theme.colors == {}
    assert theme.styles == {}
    assert not theme.description
    assert not theme.type
    assert not theme.version


def test_loads_none():
    theme = Theme.loads(None)
    assert not theme
    assert theme.definition == {}
    assert theme.colors == {}
    assert theme.styles == {}
    assert not theme.description
    assert not theme.type
    assert not theme.version


def test_description_not_defined():
    theme_str = """
    [colors]
    foreground = "#d2d2d2"
    """
    theme = Theme.loads(theme_str)
    assert not theme.description


def test_description():
    theme_str = """
    description = "my theme"
    """
    theme = Theme.loads(theme_str)
    assert theme.description == "my theme"


def test_type_not_defined():
    theme_str = """
    [colors]
    foreground = "#d2d2d2"
    """
    theme = Theme.loads(theme_str)
    assert not theme.type


def test_type():
    theme_str = """
    type = "dark"
    """
    theme = Theme.loads(theme_str)
    assert theme.type == "dark"


def test_version_not_defined():
    theme_str = """
    [colors]
    foreground = "#d2d2d2"
    """
    theme = Theme.loads(theme_str)
    assert not theme.version


def test_version():
    theme_str = """
    version = "2.1"
    """
    theme = Theme.loads(theme_str)
    assert theme.version == "2.1"


#
# test processing of colors
#
def test_color(sthm):
    assert sthm.colors["background"] == "#282a36"


def test_colors_must_be_strings():
    theme_str = """
    [colors]
    background = 282
    """
    with pytest.raises(DyeSyntaxError):
        Theme.loads(theme_str)


def test_colors_reference(sthm):
    assert sthm.colors["background_high"] == sthm.colors["background"]


def test_color_reference(sthm):
    assert sthm.colors["background_medium"] == sthm.colors["background"]


def test_colors_bare_reference(sthm):
    assert sthm.colors["background_low"] == sthm.colors["background"]


def test_colors_double1_reference(sthm):
    assert sthm.colors["background_double1"] == sthm.colors["background"]


def test_colors_double2_reference(sthm):
    assert sthm.colors["background_double2"] == sthm.colors["background"]


def test_colors_unknown_reference(sthm):
    assert sthm.colors["foreground_unknown"] == ""


def test_colors_load_order(sthm):
    assert sthm.colors["notyet"] == ""


def test_colors_subtable(sthm):
    assert isinstance(sthm.colors["triad"], dict)
    assert sthm.colors["triad.first"] == sthm.colors["purple"]


def test_colors_subtable_reference1(sthm):
    assert sthm.colors["tetrad.first"] == sthm.colors["purple"]


def test_colors_subtable_reference2(sthm):
    assert sthm.colors["tetrad.fourth"] == sthm.colors["green"]


#
# test processing of styles
#
def test_style(sthm):
    assert sthm.styles["text"].color.name == "#f8f8f2"
    assert sthm.styles["text"].color.triplet.hex == "#f8f8f2"
    assert sthm.styles["text"].bgcolor.name == "#282a36"
    assert sthm.styles["text"].bgcolor.triplet.hex == "#282a36"


def test_styles_must_be_strings():
    theme_str = """
    [styles]
    text = 282
    """
    with pytest.raises(DyeSyntaxError):
        Theme.loads(theme_str)


def test_style_color_ref(sthm):
    assert sthm.styles["foreground"].color.name == sthm.colors["foreground"]


def test_style_no_colors(sthm):
    assert sthm.styles["color1"].color.name == "#ff79c6"
    assert sthm.styles["color1"].color.triplet.hex == "#ff79c6"


def test_style_reference(sthm):
    assert sthm.styles["text_medium"] == sthm.styles["text"]


def test_styles_bare_reference(sthm):
    assert sthm.styles["text_low"] == sthm.styles["text"]


def test_styles_double1_reference(sthm):
    assert sthm.styles["text_double1"] == sthm.styles["text"]


def test_styles_double2_reference(sthm):
    assert sthm.styles["text_double2"] == sthm.styles["text"]


def test_style_unknown_reference(sthm):
    assert not sthm.styles["color2"]
    assert isinstance(sthm.styles["color2"], rich.style.Style)


def test_styles_load_order(sthm):
    assert not sthm.styles["notyet"]
    assert isinstance(sthm.styles["notyet"], rich.style.Style)


def test_style_empty(sthm):
    assert not sthm.styles["color3"]
    assert isinstance(sthm.styles["color3"], rich.style.Style)


def test_styles_subtable(sthm):
    assert isinstance(sthm.styles["triad"], dict)
    assert sthm.styles["triad.first"].color.name == "#50fa7b"


def test_styles_subtable_reference1(sthm):
    assert sthm.styles["triad.second"] == sthm.styles["foreground"]


def test_styles_subtable_reference2(sthm):
    assert sthm.styles["triad.third"].color.name == "#50fa7b"


def test_styles_subtable_reference3(sthm):
    assert sthm.styles["accent.1"].color.name == "#ffb86c"
    assert sthm.styles["accent.2"].color.name == "#ff79c6"



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""classes for storing a theme"""

import jinja2
import tomlkit
from benedict import benedict

from .utils import merge_and_process_colors, merge_and_process_styles


class Theme:
    """load and parse a toml file into a theme object"""

    # class methods to create a new theme
    @classmethod
    def loads(cls, tomlstring=None):
        """Process a given string as a theme and return a new theme object"""
        if tomlstring:  # noqa: SIM108
            toparse = tomlstring
        else:
            # tomlkit can't parse None, so if we got it as the default
            # or if the caller pased None intentionally...
            toparse = ""
        theme = Theme()
        theme.definition = benedict(tomlkit.loads(toparse))
        theme._process()
        return theme

    @classmethod
    def load(cls, fobj, filename=None):
        """Process a file object as a theme and return a new theme object

        Pass the optional filename to put in the .filename property
        of the returned theme object
        """
        theme = Theme()
        theme.definition = benedict(tomlkit.load(fobj))
        theme.filename = filename
        theme._process()
        return theme

    #
    # initialization and properties
    #
    def __init__(self):
        """Construct a new Theme object"""

        # the raw toml definition of the theme
        self.definition = benedict()

        # the processed colors, it's a dict of strings
        self.colors = benedict()

        # the processed elements, it's a dict of rich.style.Style()
        self.styles = benedict()

        # metadata from the theme file
        self.description = None
        self.type = None
        self.version = None

        # a place to stash the file that the theme was loaded from
        # it's up to the caller/user to make sure this is set properly
        # defaults to None
        self.filename = None

    def __bool__(self):
        """Falsy if the definition is falsy, truthy if the definition is truthy"""
        return bool(self.definition)

    def _process(self):
        """process a newly loaded definition, including variables and styles

        this sets self.colors and self.styles
        """

        self.description = self.definition.get("description", None)
        self.type = self.definition.get("type", None)
        self.version = self.definition.get("version", None)

        env = jinja2.Environment()

        try:
            raw_colors = benedict(self.definition["colors"])
        except KeyError:
            raw_colors = benedict()
        self.colors = benedict()
        merge_and_process_colors(self.colors, raw_colors, env)

        # process the styles, using the colors as variables
        # each style in should be a rich.Style() object
        try:
            raw_styles = benedict(self.definition["styles"])
        except KeyError:
            raw_styles = benedict()
        self.styles = benedict()
        merge_and_process_styles(self.styles, raw_styles, env, self.colors)



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""eza agent for EZA_COLORS"""

import contextlib

from ..exceptions import DyeSyntaxError
from .base import AgentBase, LsColorsFromStyle


class Eza(AgentBase, LsColorsFromStyle):
    "Create EZA_COLORS environment variable for use with ls replacement eza"

    EZA_COLORS_BASE_MAP = {
        # create a map of friendly names to real codes.
        # we use the same friendly names as the theme.yml
        # file used by eza
        #
        "filekinds:normal": "fi",
        "filekinds:directory": "di",
        "filekinds:symlink": "ln",
        "filekinds:pipe": "pi",
        "filekinds:block_device": "bd",
        "filekinds:char_device": "cd",
        "filekinds:socket": "so",
        "filekinds:special": "sp",
        "filekinds:executable": "ex",
        "filekinds:mount_point": "mp",
        #
        "perms:user_read": "ur",
        "perms:user_write": "uw",
        "perms:user_executable_file": "ux",
        "perms:user_execute_other": "ue",
        "perms:group_read": "gr",
        "perms:group_write": "gw",
        "perms:group_execute": "gx",
        "perms:other_read": "tr",
        "perms:other_write": "tw",
        "perms:other_execute": "tx",
        "perms:special_user_file": "su",
        "perms:special_other": "sf",
        "perms:attribute": "xa",
        #
        "size:major": "df",
        "size:minor": "ds",
        "size:number_style": "sn",
        "size:number_byte": "nb",
        "size:number_kilo": "nk",
        "size:number_mega": "nm",
        "size:number_giga": "ng",
        "size:number_huge": "nt",
        "size:unit_style": "sb",
        "size:unit_byte": "ub",
        "size:unit_kilo": "uk",
        "size:unit_mega": "um",
        "size:unit_giga": "ug",
        "size:unit_huge": "ut",
        #
        "users:user_you": "uu",
        "users:user_other": "un",
        "users:user_root": "uR",
        "users:group_yours": "gu",
        "users:group_other": "gn",
        "users:group_root": "gR",
        #
        "links:normal": "lc",
        "links:multi_link_file": "lm",
        #
        "git:new": "ga",
        "git:modified": "gm",
        "git:deleted": "gd",
        "git:renamed": "gv",
        "git:typechange": "gt",
        "git:ignored": "gi",
        "git:conflicted": "gc",
        #
        "git_repo:branch_main": "Gm",
        "git_repo:branch_other": "Go",
        "git_repo:git_clean": "Gc",
        "git_repo:git_dirty": "Gd",
        #
        "selinux:colon": "Sn",
        "selinux:user": "Su",
        "selinux:role": "Sr",
        "selinux:typ": "St",
        "selinux:range": "Sl",
        #
        "file_type:image": "im",
        "file_type:video": "vi",
        "file_type:music": "mu",
        "file_type:lossless": "lo",
        "file_type:crypto": "cr",
        "file_type:document": "do",
        "file_type:compressed": "co",
        "file_type:temp": "tm",
        "file_type:compiled": "cm",
        "file_type:build": "bu",
        "file_type:source": "sc",
        #
        "punctuation": "xx",
        "date": "da",
        "inode": "in",
        "blocks": "bl",
        "header": "hd",
        "octal": "oc",
        "flags": "ff",
        "symlink_path": "lp",
        "control_char": "cc",
        "broken_path_overlay": "b0",
        "broken_symlink": "or",
    }
    # this map allows you to either use the 'native' eza code, or the
    # 'friendly' name defined by shell-themer
    EZA_COLORS_MAP = {}
    for friendly, actual in EZA_COLORS_BASE_MAP.items():
        EZA_COLORS_MAP[friendly] = actual
        EZA_COLORS_MAP[actual] = actual

    def run(self, comments=False):
        "Render a EZA_COLORS variable suitable for eza"
        outlist = []
        # figure out if we are clearing builtin styles
        with contextlib.suppress(KeyError):
            clear_builtin = self.scope.definition["clear_builtin"]
            if not isinstance(clear_builtin, bool):
                raise DyeSyntaxError(
                    f"scope '{self.scope}' requires 'clear_builtin' to be true or false"
                )
            if clear_builtin:
                # this tells exa to not use any built-in/hardcoded colors
                outlist.append("reset")

        # iterate over the styles given in our configuration
        for name, style in self.scope.styles.items():
            if style:
                _, render = self.ls_colors_from_style(
                    name,
                    style,
                    self.EZA_COLORS_MAP,
                    allow_unknown=True,
                    scope_name=self.scope,
                )
                outlist.append(render)

        # process the filesets

        # figure out which environment variable to put it in
        try:
            varname = self.scope.definition["environment_variable"]
        except KeyError:
            varname = "EZA_COLORS"

        # even if outlist is empty, we have to set the variable, because
        # when we are switching a theme, there may be contents in the
        # environment variable already, and we need to tromp over them
        # we chose to set the variable to empty instead of unsetting it
        print(f'''export {varname}="{":".join(outlist)}"''')



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""utility functions and classes"""

from enum import Enum
from importlib import metadata

import benedict
import rich

from .exceptions import DyeSyntaxError


class DefinitionSource(Enum):
    """Enum to indicate where a style or color is defined"""

    def __str__(self):
        return self.value

    THEME = "theme"
    PATTERN = "pattern"


def version_string():
    """return a version string suitable for display to a user"""
    try:
        # this is the distribution name, not the package name
        ver = metadata.version("dye-my-shell")
    except metadata.PackageNotFoundError:  # pragma: nocover
        ver = "unknown"
    return ver


def merge_and_process_colors(base_colors, merging_colors, jinja_env):
    """merge together two benedicts of colors (one or both can be empty) and
    process references and templates in all the values.

    Args:
        base_colors (benedict): _description_
        merging_colors (benedict): _description_
        jinja_env (jinja2.environments): _description_

    base_colors will be modified to include all of the items in merging_colors.

    items in merging_colors will overwrite items already in base_colors.

    Used by Theme and Pattern objects

    This assumes that base_colors (if present) has already been processed. If
    you just need to process one dict, call this like

        processed_dict = benedict()
        merge_and_process_colors(processed_dict, unprocessed_dict)

    that will result in processed_dict having all the keys of unprocessed_dict
    but all the values will have been processed
    """
    # iterate over all the keys in raw_colors, processing and inserting
    # the values into self.colors
    keylist = merging_colors.keypaths(sort=False)
    for key in keylist:
        value = merging_colors[key]
        if isinstance(value, str):
            if value in base_colors:
                # bare lookup so that foreground_low = "foreground" works
                base_colors[key] = base_colors[value]
            else:
                template = jinja_env.from_string(value)
                base_colors[key] = template.render(
                    # this lets us do {{colors.foreground}} or {{color.foreground}}
                    colors=base_colors,
                    color=base_colors,
                )
        elif isinstance(value, dict):
            # So we rely on and test for the fact that both the subtable name
            # (with the dict value) and each of the subtable keys show up in
            # keylist. Those other keys will be picked up in a different iteration
            # of the loop. But they shouldn't be a syntax error.
            pass
        else:
            raise DyeSyntaxError(f"color {key} must be defined as a string")


def merge_and_process_styles(base_styles, merging_styles, jinja_env, colors=None):
    """Merge together two benedicts of colors (one or both can be empty) and
    process references and templates in all the values.

    Args:
        base_styles (benedict): _description_
        merging_styles (benedict): _description_
        jinja_env (jinja2.environments): _description_
        colors (benedicty): optional benedict of processed colors

    base_styles will be modified to include all of the items in merging_styles.

    items in merging_styles will overwrite items already in base_styles.

    Used by Theme and Pattern objects

    This assumes that base_styles (if present) has already been processed. If
    you just need to process one dict, call this like

        processed_dict = benedict()
        merge_and_process_styles(processed_dict, unprocessed_dict)

    that will result in processed_dict having all the keys of unprocessed_dict
    but all the values will have been processed
    """
    if not colors:
        colors = benedict.benedict()

    keylist = merging_styles.keypaths(sort=False)
    for key in keylist:
        value = merging_styles[key]
        if isinstance(value, str):
            if value in base_styles:
                # bare lookup so that foreground_low = "foreground" works
                base_styles[key] = base_styles[value]
            else:
                template = jinja_env.from_string(value)
                rendered = template.render(
                    # allow {{style.foreground}} or {{styles.foreground}}
                    # or {{color.background}} or {{colors.background}}
                    color=colors,
                    colors=colors,
                    style=base_styles,
                    styles=base_styles,
                )
                base_styles[key] = rich.style.Style.parse(rendered)
        elif isinstance(value, dict):
            # So we rely on and test for the fact that both the subtable name
            # (with the dict value) and each of the subtable keys show up in
            # keylist. Those other keys will be picked up in a different iteration
            # of the loop. But they shouldn't be a syntax error.
            pass
        else:
            raise DyeSyntaxError(f"theme {key} must be defined as a string")



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""class for storing and processing a scope"""

import contextlib
import subprocess

import jinja2
import rich
import rich.text
from benedict import benedict

from .agents import AgentBase
from .exceptions import DyeError, DyeSyntaxError
from .filters import jinja_filters


class Scope:
    """represent a scope

    This class handles stuff that is the same for all scopes,
    no matter what the agent for that scope is. This includes:

        - making sure an agent is defined
        - instantiating the agent class
        - processing templates (colors, styles, variables) in the scope
        - processing styles in the scope

    When you want to run this scope, ie have the agent generate output
    based on the scope, you should do:

        scope.run_agent()
    """

    COMMENT_BEGIN = "# "

    #
    # initialization and properties
    #
    def __init__(self, name, pattern):
        """Construct a new Scope object"""

        self.name = name
        """Name of the scope, pass it to the constructor"""

        self.definition = {}
        """A toml definition of the scope, which has had all templates processed"""

        self.styles = {}
        """The dictionary of style objects defined in this scope"""

        self.agent_name = None
        self.agent = None

        self._process(pattern, name)

    def _process(self, pattern, name):
        """Process the scope definition

        this sets self.definition, which contains
        toml that has had all templates (i.e. colors, styles, variables)
        processed in all values

        it also sets self.styles, as a dict of style objects
        """

        self.name = name

        env = jinja2.Environment()
        env.filters = jinja_filters()
        data = {}
        data["color"] = pattern.colors
        data["colors"] = pattern.colors
        data["style"] = pattern.styles
        data["styles"] = pattern.styles
        data["var"] = pattern.variables
        data["vars"] = pattern.variables
        data["variable"] = pattern.variables
        data["variables"] = pattern.variables
        env.globals = data

        def render_func(d, key, value):
            # only process strings
            if isinstance(value, str):
                template = env.from_string(value)
                d[key] = template.render()

        try:
            scopedef = benedict(pattern.definition["scopes"][name])
        except KeyError as exc:
            raise DyeError(f"{name}: no such scope") from exc
        scopedef.traverse(render_func)
        self.definition = scopedef

        self._process_agent()
        self._process_scope_styles(pattern)

    def _process_agent(self):
        """validate and set self.agent_name and self.agent"""
        try:
            self.agent_name = self.definition["agent"]
        except KeyError:
            errmsg = f"scope '{self.name}' does not have an agent."
            raise DyeSyntaxError(errmsg) from None
        try:
            # go get the apropriate class for the agent
            agent_cls = AgentBase.classmap[self.agent_name]
            # initialize the class with the scope (that's our self)
            self.agent = agent_cls(self)
        except KeyError as exc:
            raise DyeError(
                f"{self.agent_name}: unknown agent in scope '{self.name}"
            ) from exc

    def _process_scope_styles(self, pattern):
        """create a dictionary of style objects parsed from self.definition

        sets self.styles
        """
        try:
            raw_styles = self.definition["style"]
        except (KeyError, TypeError):
            raw_styles = {}
        try:
            more_styles = self.definition["styles"]
        except (KeyError, TypeError):
            more_styles = {}
        # merge these two together, the order we do this means that
        # styles override style
        for name, styledef in more_styles.items():
            raw_styles[name] = styledef

        processed_styles = {}
        for name, styledef in raw_styles.items():
            # lookup the style by name in pattern.styles
            if styledef in pattern.styles:
                processed_styles[name] = pattern.styles[styledef]
            else:
                processed_styles[name] = rich.style.Style.parse(styledef)

        self.styles = processed_styles

    def run_agent(self, comments=False, console=None, output_elements=None):
        """
        returns output consisting of shell commands which must
        be sourced in the current shell in order to become active
        """
        if self._enabled():
            if comments:
                self._print_comment(f"scope '{self.name}'", console, output_elements)
            # run the agent, printing any shell commands it returns
            output = self.agent.run(comments)
            if output:
                print(output)
        else:
            if comments:
                self._print_comment(
                    f"scope '{self.name}' skipped because it is not enabled",
                    console,
                    output_elements,
                )

    def _print_comment(self, msg, console=None, output_elements=None):
        """print a comment line with optional styling"""
        if console and output_elements:
            text = rich.text.Text()
            text.append(
                self.COMMENT_BEGIN,
                style=output_elements.get("comment_begin"),
            )
            text.append(msg, style=output_elements.get("comment_text"))
            console.print(text)
        else:
            print(f"{self.COMMENT_BEGIN}{msg}")

    def _enabled(self):
        """Determine if the scope is enabled
        The default is that the scope is enabled

        returns whether this scope is enabled

        If can be disabled by any of these in the pattern file:

            enabled = false

        or:
            enabled_if = "{shell cmd}" returns a non-zero exit code

        if 'enabled = false' is present, then enabled_if is not checked

        May print output to be executed by the shell, may execute
        shell commands too
        """
        with contextlib.suppress(KeyError):
            enabled = self.definition["enabled"]
            if not isinstance(enabled, bool):
                raise DyeSyntaxError(
                    f"scope '{self.name}' requires 'enabled' to be true or false"
                )
            # this is authoritative, if it exists, ignore enabled_if below
            return enabled

        # no enabled directive, so we check for enabled_if
        try:
            enabled_if = self.definition["enabled_if"]
            if not enabled_if:
                # we have a key, but an empty value (aka command)
                # by rule we say it's enabled
                return True
        except KeyError:
            # no enabled_if key, so we must be enabled
            return True

        # if we get here we have something in enabled_if that
        # we need to go run
        proc = subprocess.run(enabled_if, shell=True, check=False, capture_output=True)
        if proc.returncode != 0:  # noqa: SIM103
            # the shell command returned a non-zero exit code
            # and this scope should therefore be disabled
            return False
        return True



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

import os

import pytest

from dye import Dye

#
# test scope selection
#
SCOPE_PATTERN = """
    [colors]
    background =  "#282a36"
    foreground =  "#f8f8f2"

    [styles]
    current_line =  "#f8f8f2 on #44475a"
    comment =  "#6272a4"
    cyan =  "#8be9fd"
    green =  "#50fa7b"
    orange =  "#ffb86c"
    pink =  "#ff79c6"
    purple =  "#bd93f9"
    red =  "#ff5555"
    yellow =  "#f1fa8c"

    [scopes.iterm]
    agent = "iterm"
    styles.foreground = "{{colors.foreground}}"
    styles.background = "{{colors.background}}"

    [scopes.fzf]
    agent = "fzf"
    environment_variable = "FZF_DEFAULT_OPTS"

    opt.--prompt = ">"
    opt.--border = "single"
    opt.--pointer = "•"
    opt.--info = "hidden"
    opt.--no-sort = true
    opt."+i" = true

    styles.text = "{{colors.foreground}}"
    styles.label = "{{styles.green}}"
    styles.border = "{{styles.orange}}"
    styles.selected = "{{styles.current_line}}"
    styles.prompt = "{{styles.green}}"
    styles.indicator = "{{styles.cyan}}"
    styles.match = "{{styles.pink}}"

    [scopes.env]
    agent = "environment_variables"
    unset = "NO_COLOR"
    """


def test_single_scope(dye_cmdline, capsys):
    exit_code = dye_cmdline("apply -s fzf", None, SCOPE_PATTERN)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert out
    assert not err
    lines = out.splitlines()
    assert len(lines) == 1
    assert "export FZF_DEFAULT_OPTS=" in lines[0]


def test_multiple_scopes(dye_cmdline, capsys):
    exit_code = dye_cmdline("apply -s env,fzf", None, SCOPE_PATTERN)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert out
    assert not err
    lines = out.splitlines()
    assert len(lines) == 2
    # scopes should come out in the order specified
    assert "unset NO_COLOR" in lines[0]
    assert "export FZF_DEFAULT_OPTS" in lines[1]


def test_all_scopes(dye_cmdline, capsys):
    exit_code = dye_cmdline("apply", None, SCOPE_PATTERN)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert out
    assert not err
    lines = out.splitlines()
    # 2 lines from iterm, 1 from fzf, 1 for env
    assert len(lines) == 4


def test_unknown_scope(dye_cmdline, capsys):
    exit_code = dye_cmdline("apply -s unknownscope", None, SCOPE_PATTERN)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert err == "dye: unknownscope: no such scope\n"


def test_no_scopes(dye_cmdline, capsys):
    pattern_str = """
        [styles]
        background =  "#282a36"
        foreground =  "#f8f8f2"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not out
    assert not err


#
# test enabled
#
def test_enabled(dye_cmdline, capsys):
    pattern_str = """
        [scopes.one]
        enabled = false
        agent = "environment_variables"
        unset = "NOLISTVAR"

        [scopes.two]
        enabled = true
        agent = "environment_variables"
        unset = "SOMEVAR"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert "unset SOMEVAR" in out
    assert "unset NOLISTVAR" not in out


def test_enabled_false_enabled_if_ignored(dye_cmdline, capsys):
    pattern_str = """
        [scopes.unset]
        enabled = false
        enabled_if = "[[ 1 == 1 ]]"
        agent = "environment_variables"
        unset = "NOLISTVAR"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert not out


def test_enabled_true_enabed_if_ignored(dye_cmdline, capsys):
    pattern_str = """
        [scopes.unset]
        enabled = true
        enabled_if = "[[ 0 == 1 ]]"
        agent = "environment_variables"
        unset = "NOLISTVAR"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert "unset NOLISTVAR" in out


def test_enabled_invalid_value(dye_cmdline, capsys):
    pattern_str = """
        [scopes.unset]
        enabled = "notaboolean"
        agent = "environment_variables"
        unset = "NOLISTVAR"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_ERROR
    assert not out
    assert "to be true or false" in err


ENABLED_IFS = [
    ("", True),
    ("echo", True),
    ("[[ 1 == 1 ]]", True),
    ("[[ 1 == 0 ]]", False),
    ("{{variables.echocmd}} hi", True),
    ("{{variables.falsetest}}", False),
]


@pytest.mark.parametrize("cmd, enabled", ENABLED_IFS)
def test_activate_enabled_if(cmd, enabled, dye_cmdline, capsys):
    pattern_str = f"""
        [variables]
        echocmd = "builtin echo"
        falsetest = "[[ 1 == 0 ]]"

        [scopes.unset]
        enabled_if = "{cmd}"
        agent = "environment_variables"
        unset = "ENVVAR"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    if enabled:
        assert "unset ENVVAR" in out
    else:
        assert not out


#
# test comments
#
def test_comments(dye_cmdline, capsys):
    pattern_str = """
        [scopes.nolistvar]
        enabled = false
        agent = "environment_variables"
        unset = "NOLISTVAR"

        [scopes.somevar]
        enabled = true
        agent = "environment_variables"
        unset = "SOMEVAR"
    """
    exit_code = dye_cmdline("apply --comment", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    lines = out.splitlines()
    # there is a long message on skipped
    assert "# scope 'nolistvar' skipped" in out
    assert "# scope 'somevar'" in lines
    assert "unset SOMEVAR" in lines
    assert "unset NOLISTVAR" not in lines


def test_comments_styled(dye_cmdline, mocker, capsys):
    mocker.patch.dict(
        os.environ,
        {"DYE_COLORS": "comment_begin=red:comment_text=#00ff00"},
    )
    pattern_str = """
        [scopes.nolistvar]
        enabled = false
        agent = "environment_variables"
        unset = "NOLISTVAR"

        [scopes.somevar]
        enabled = true
        agent = "environment_variables"
        unset = "SOMEVAR"
    """
    exit_code = dye_cmdline("apply --comment", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert "scope 'nolistvar' skipped" in out
    assert "scope 'somevar'" in out
    assert "unset SOMEVAR" in out
    # ANSI escape codes should be present because styles are applied
    assert "\x1b[" in out



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

from dye import Dye


def test_dye_single_style(dye_cmdline, capsys):
    pattern_str = """
        [scopes.my_dye]
        agent = "dye"
        style.usage_metavar = "#ffb86c"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out.strip() == 'export DYE_COLORS="usage_metavar=#ffb86c"'


def test_dye_multiple_styles(dye_cmdline, capsys):
    pattern_str = """
        [scopes.my_dye]
        agent = "dye"
        style.usage_metavar = "#ffb86c"
        styles.ui_border = "#bd93f9"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert "usage_metavar=#ffb86c" in out
    assert "ui_border=#bd93f9" in out
    assert out.strip().startswith('export DYE_COLORS="')
    assert out.strip().endswith('"')


def test_dye_with_template(dye_cmdline, capsys):
    pattern_str = """
        [colors]
        orange = "#ffb86c"
        purple = "#bd93f9"

        [scopes.my_dye]
        agent = "dye"
        style.usage_metavar = "{{ colors.orange }}"
        style.ui_border = "{{ colors.purple }}"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert "usage_metavar=#ffb86c" in out
    assert "ui_border=#bd93f9" in out


def test_dye_with_bold_style(dye_cmdline, capsys):
    pattern_str = """
        [scopes.my_dye]
        agent = "dye"
        style.usage_prog = "bold #50fa7b"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out.strip() == 'export DYE_COLORS="usage_prog=bold #50fa7b"'


def test_dye_no_styles(dye_cmdline, capsys):
    pattern_str = """
        [scopes.my_dye]
        agent = "dye"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    assert out.strip() == 'export DYE_COLORS=""'


def test_dye_all_elements(dye_cmdline, capsys):
    pattern_str = """
        [scopes.my_dye]
        agent = "dye"
        style.usage_args = "#f8f8f2"
        style.usage_groups = "bold #ff5555"
        style.usage_help = "#6272a4"
        style.usage_metavar = "#ffb86c"
        style.usage_prog = "bold #50fa7b"
        style.usage_syntax = "#f1fa8c"
        style.usage_text = "#f8f8f2"
        style.ui_border = "#bd93f9"
        style.ui_column_header = "bold #ff79c6"
        style.error_progname = "bold #ff5555"
        style.error_text = "#ff5555"
        style.debug_label = "bold #6272a4"
        style.debug_text = "#6272a4"
        style.comment_begin = "bold #6272a4"
        style.comment_text = "#6272a4"
    """
    exit_code = dye_cmdline("apply", None, pattern_str)
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    for element in Dye.OUTPUT_ELEMENTS:
        assert f"{element}=" in out



#
# Copyright (c) 2025 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# pylint: disable=protected-access, missing-function-docstring, redefined-outer-name
# pylint: disable=missing-module-docstring, unused-variable

from dye import Dye


#
# test the agents command (which shows the list of available agents)
#
def test_agents(dye_cmdline, capsys):
    exit_code = dye_cmdline("agents")
    out, err = capsys.readouterr()
    assert exit_code == Dye.EXIT_SUCCESS
    assert not err
    expected_agents = [
        "dye",
        "environment_variables",
        "eza",
        "fzf",
        "iterm",
        "ls_colors",
        "shell",
    ]
    for agent in expected_agents:
        assert agent in out
    # the agents command outputs a rich table with a header row and a line
    # under the header
    lines = out.splitlines()
    # this is the header row
    assert "Agent" in lines[1]
    # +4 for the top border, bottom border, header words, and line under
    # the header words
    assert len(lines) == len(expected_agents) + 4



#
# Copyright (c) 2023 Jared Crapo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""dye agent for setting DYE_COLORS

dye dyes itself using this dye agent
"""

from .base import AgentBase


class Dye(AgentBase):
    "Set DYE_COLORS environment variable for dye's own color output"

    def run(self, comments=False) -> str:
        outlist = []
        for name, styledef in self.scope.styles.items():
            # str() turns the style definition into a plain text string
            # resolving all the themes/patterns/references, etc
            outlist.append(f"{name}={str(styledef)}")
        return f'export DYE_COLORS="{":".join(outlist)}"'
